# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
An immutable list view.
"""

from __future__ import print_function, absolute_import

try:
    from collections.abc import Sequence
except ImportError:
    # Python 2
    from collections import Sequence  # pylint: disable=deprecated-class

__all__ = ['ListView']


class ListView(Sequence):
    # pylint: disable=line-too-long
    """
    An immutable list view.

    Derived from :class:`~py3:collections.abc.Sequence`.

    This class provides an immutable view on a possibly mutable sequence
    object. The sequence object must be an instance of
    :class:`~py3:collections.abc.Sequence`, e.g. :class:`list`, :class:`tuple`,
    :class:`range`, or a user-defined class.

    This can be used for example when a class maintains a list that should
    be made available to users of the class without allowing them to modify the
    list.

    In the description of this class, the term 'view' always refers to the
    :class:`ListView` object, and the term 'list' or 'underlying list' refers
    to the sequence object the view is based on.

    The :class:`ListView` class supports the complete behavior of Python class
    :class:`list`, except for any methods that would modify the list.
    Note that the non-modifying methods of class :class:`list` are a superset of
    the methods defined for the abstract class
    :class:`~py3:collections.abc.Sequence` (the methods are listed in the table
    at the top of the linked page).

    The view is "live": Since the view class delegates all operations to the
    underlying list, any modification of the underlying list object
    will be visible in the view object.

    Note that only the view object is immutable, not its items. So if the values
    in the underlying list are mutable objects, they can be modified through the
    view.

    Note that in Python, augmented assignment (e.g. ``x += y``) is not
    guaranteed to modify the left hand object in place, but can result in the
    left hand name being bound to a new object (like in ``x = x + y``).
    For details, see
    `object.__iadd__() <https://docs.python.org/3/reference/datamodel.html#object.__iadd__>`_.

    For the ListView class, augmented assignment is supported and results in
    binding the left hand name to a new ListView object.
    """  # noqa: E501
    # pylint: enable=line-too-long

    __slots__ = ['_list']

    def __init__(self, a_list):
        """
        Parameters:

          a_list (:class:`~py3:collections.abc.Sequence`):
            The underlying list.
            If this object is a ListView, its underlying list is used.
        """
        if not isinstance(a_list, Sequence):
            raise TypeError(
                "The a_list parameter must be a Sequence, but is: {}".
                format(type(a_list)))
        if isinstance(a_list, ListView):
            a_list = a_list._list
        self._list = a_list

    @property
    def list(self):
        """
        The underlying list.

        Access to the underlying list is provided for purposes such as
        conversion to JSON or other cases where the view classes do not work.
        This access should not be used to modify the underlying list.
        """
        return self._list

    def __repr__(self):
        """
        ``repr(self)``:
        Return a string representation of the view suitable for debugging.

        The underlying list is represented using its ``repr()``
        representation.
        """
        return "{0.__class__.__name__}({1!r})".format(self, self._list)

    def __getstate__(self):
        """Support for pickling."""
        # Needed on Python 2 due to the use of slots
        a_dict = dict()
        for attr in self.__slots__:
            a_dict[attr] = getattr(self, attr)

        # Support for objects that also have __dict__, e.g. user defined
        # derived classes that did not define __slots__:
        if hasattr(self, '__dict__'):
            for attr in self.__dict__:
                a_dict[attr] = getattr(self, attr)

        return a_dict

    def __setstate__(self, a_dict):
        """Support for unpickling."""
        # Needed on Python 2 due to the use of slots
        for attr in a_dict:
            setattr(self, attr, a_dict[attr])

    def __getitem__(self, index):
        """
        ``self[index]``:
         Return the list value at the index position.

        Raises:
          IndexError: Index out of range.
        """
        return self._list[index]

    def __len__(self):
        """
        ``len(self)``:
        Return the number of items in the list.

        The return value is the number of items in the underlying list.
        """
        return len(self._list)

    def __contains__(self, value):
        """
        ``value in self``:
        Return a boolean indicating whether the list contains a value.

        The return value indicates whether the underlying list contains an
        item that is equal to the value.
        """
        return value in self._list

    def __iter__(self):
        """
        Return an iterator through the list items.
        """
        return iter(self._list)

    def __add__(self, other):
        """
        ``self + other``:
        Return a new view on the concatenation of the list and the other list.

        The returned :class:`ListView` object is a view on a new list object of
        the type of the left hand operand that contains the items that are in
        the underlying list of the left hand operand, concatenated with the
        items in the other list (or in case of a ListView, its underlying list).

        The other object must be an :term:`py3:iterable` or :class:`ListView`.

        The list and the other list are not changed.

        Raises:
          TypeError: The other object is not an iterable.
        """
        # pylint: disable=protected-access
        other_list = other._list if isinstance(other, ListView) else other
        new_list = self._list + other_list
        return ListView(new_list)

    def __mul__(self, number):
        """
        ``self * number``:
        Return a new view on the multiplication of the list with a number.

        The returned :class:`ListView` object is a view on a new list object of
        the type of the left hand operand that contains the items that are in
        the underlying list of the left hand operand as many times as specified
        by the right hand operand.

        A number <= 0 causes the returned list to be empty.

        The left hand operand is not changed.
        """
        new_list = self._list * number
        return ListView(new_list)

    def __rmul__(self, number):
        """
        ``number * self``:
        Return a new view on the multiplication of the list with a number.

        This method is a fallback and is called only if the left operand does
        not support the operation.

        The returned :class:`ListView` object is a view on a new list object of
        the type of the right hand operand that contains the items that are in
        the underlying list of the right hand operand as many times as specified
        by the left hand operand.

        A number <= 0 causes the returned list to be empty.

        The right hand operand is not changed.
        """
        return self * number  # Delegates to __mul__()

    def __reversed__(self):
        """
        ``reversed(self) ...``:
        Return an iterator through the list in reversed iteration order.

        The returned iterator yields the items in the underlying list in the
        reversed iteration order.
        """
        return reversed(self._list)

    def __eq__(self, other):
        """
        ``self == other``:
        Return a boolean indicating whether the list is equal to the other list.

        The return value indicates whether the items in the underlying list are
        equal to the items in the other list (or in case of a ListView, its
        underlying list).

        The other object must be a :class:`list` or :class:`ListView`.

        Raises:
          TypeError: The other object is not a list or ListView.
        """
        # pylint: disable=protected-access
        other_list = other._list if isinstance(other, ListView) else other
        return self._list == other_list

    def __ne__(self, other):
        """
        ``self != other``:
        Return a boolean indicating whether the list is not equal to the other
        list.

        The return value indicates whether the items in the underlying list are
        not equal to the items in the other list (or in case of a ListView, its
        underlying list).

        The other object must be a :class:`list` or :class:`ListView`.

        Raises:
          TypeError: The other object is not a list or ListView.
        """
        # pylint: disable=protected-access
        other_list = other._list if isinstance(other, ListView) else other
        return self._list != other_list

    def __gt__(self, other):
        # pylint: disable=line-too-long
        """
        ``self > other``:
        Return a boolean indicating whether the list is greater than the other
        list.

        The return value indicates whether the underlying list is greater than
        the other list (or in case of a ListView, its underlying list), based on
        the lexicographical ordering Python defines for sequence types
        (see https://docs.python.org/3/tutorial/datastructures.html#comparing-sequences-and-other-types)

        The other object must be a :class:`list` or :class:`ListView`.

        Raises:
          TypeError: The other object is not a list or ListView.
        """  # noqa: E501
        # pylint: enable=line-too-long
        # pylint: disable=protected-access
        other_list = other._list if isinstance(other, ListView) else other
        return self._list > other_list

    def __lt__(self, other):
        # pylint: disable=line-too-long
        """
        ``self < other``:
        Return a boolean indicating whether the list is less than the other
        list.

        The return value indicates whether the underlying list is less than
        the other list (or in case of a ListView, its underlying list), based on
        the lexicographical ordering Python defines for sequence types
        (see https://docs.python.org/3/tutorial/datastructures.html#comparing-sequences-and-other-types)

        The other object must be a :class:`list` or :class:`ListView`.

        Raises:
          TypeError: The other object is not a list or ListView.
        """  # noqa: E501
        # pylint: enable=line-too-long
        # pylint: disable=protected-access
        other_list = other._list if isinstance(other, ListView) else other
        return self._list < other_list

    def __ge__(self, other):
        # pylint: disable=line-too-long
        """
        ``self < other``:
        Return a boolean indicating whether the list is greater than or equal to
        the other list.

        The return value indicates whether the underlying list is greater than
        or equal to the other list (or in case of a ListView, its underlying
        list), based on the lexicographical ordering Python defines for sequence
        types
        (see https://docs.python.org/3/tutorial/datastructures.html#comparing-sequences-and-other-types)

        The other object must be a :class:`list` or :class:`ListView`.

        Raises:
          TypeError: The other object is not a list or ListView.
        """  # noqa: E501
        # pylint: enable=line-too-long
        # pylint: disable=protected-access
        other_list = other._list if isinstance(other, ListView) else other
        return self._list >= other_list

    def __le__(self, other):
        # pylint: disable=line-too-long
        """
        ``self < other``:
        Return a boolean indicating whether the list is less than or equal to
        the other list.

        The return value indicates whether the underlying list is less than or
        equal to the other list (or in case of a ListView, its underlying list),
        based on the lexicographical ordering Python defines for sequence types
        (see https://docs.python.org/3/tutorial/datastructures.html#comparing-sequences-and-other-types)

        The other object must be a :class:`list` or :class:`ListView`.

        Raises:
          TypeError: The other object is not a list or ListView.
        """  # noqa: E501
        # pylint: enable=line-too-long
        # pylint: disable=protected-access
        other_list = other._list if isinstance(other, ListView) else other
        return self._list <= other_list

    def count(self, value):
        """
        Return the number of times the specified value occurs in the list.
        """
        return self._list.count(value)

    def copy(self):
        """
        Return a new view on a shallow copy of the list.

        The returned :class:`ListView` object is a new view object on a list
        object of the type of the underlying list.

        If the list type is immutable, the returned list object may be the
        underlying list object. If the list type is mutable, the returned list
        is a new list object that is a shallow copy of the underlying list
        object.
        """
        org_class = self._list.__class__
        new_list = org_class(self._list)  # May be same object if immutable
        return ListView(new_list)

    def index(self, value, start=0, stop=9223372036854775807):
        """
        Return the index of the first item in the list with the specified value.

        The search is limited to the index range defined by the specified
        ``start`` and ``stop`` parameters, whereby ``stop`` is the index
        of the first item after the search range.

        Raises:
          ValueError: No such item is found.
        """
        return self._list.index(value, start, stop)

    def __hash__(self):
        """
        ``hash(self)``:
        Return a hash value for the list.

        Whether hashing is supported depends on the underlying list. For
        example, the standard Python :class:`list` class does not support
        hashing, but the standard Python :class:`tuple` class does.

        Raises:
          TypeError: The underlying list does not support hashing.
        """
        return hash(self._list)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
An immutable dictionary view.
"""

from __future__ import print_function, absolute_import

import sys
import os
try:
    from collections.abc import Mapping
except ImportError:
    # Python 2
    from collections import Mapping  # pylint: disable=deprecated-class

__all__ = ['DictView']

# This env var is set when building the docs. It causes the methods that are
# supposed to exist only under certain circumstances, not to be removed, so
# that they appear in the docs.
_BUILDING_DOCS = os.environ.get('BUILDING_DOCS', False)

# Indicates Python dict supports the iter..() and view..() methods
_DICT_SUPPORTS_ITER_VIEW = sys.version_info[0:2] == (2, 7)

# Indicates Python dict supports the has_key() method
_DICT_SUPPORTS_HAS_KEY = sys.version_info[0:2] == (2, 7)

# Indicates Python dict supports the __reversed__() method
_DICT_SUPPORTS_REVERSED = sys.version_info[0:2] >= (3, 8)

# Indicates Python dict supports the __or__/__ror__() methods
_DICT_SUPPORTS_OR = sys.version_info[0:2] >= (3, 9)


class DictView(Mapping):
    # pylint: disable=line-too-long
    """
    An immutable dictionary view.

    Derived from :class:`~py3:collections.abc.Mapping`.

    This class provides an immutable view on a possibly mutable mapping
    object. The mapping object must be an instance of
    :class:`~py3:collections.abc.Mapping`, e.g. :class:`dict`, or a
    user-defined class.

    This can be used for example when a class maintains a dictionary that should
    be made available to users of the class without allowing them to modify the
    dictionary.

    In the description of this class, the term 'view' always refers to the
    :class:`DictView` object, and the term 'dictionary' or
    'underlying dictionary' refers to the mapping object the view is based on.

    The :class:`DictView` class supports the complete behavior of Python class
    :class:`dict`, except for any methods that would modify the dictionary.
    Note that the non-modifying methods of class :class:`dict` are a superset of
    the methods defined for the abstract class
    :class:`~py3:collections.abc.Mapping` (the methods are listed in the table
    at the top of the linked page).

    The view is "live": Since the view class delegates all operations to the
    underlying dictionary, any modification of the underlying dictionary object
    will be visible in the view object.

    Note that only the view object is immutable, not its items. So if the values
    in the underlying dictionary are mutable objects, they can be modified
    through the view.

    Note that in Python, augmented assignment (e.g. ``x += y``) is not
    guaranteed to modify the left hand object in place, but can result in the
    left hand name being bound to a new object (like in ``x = x + y``).
    For details, see
    `object.__iadd__() <https://docs.python.org/3/reference/datamodel.html#object.__iadd__>`_.

    For the DictView class, augmented assignment is supported and results in
    binding the left hand name to a new DictView object.
    """  # noqa: E501
    # pylint: enable=line-too-long

    __slots__ = ['_dict']

    def __init__(self, a_dict):
        """
        Parameters:

          a_dict (:class:`~py3:collections.abc.Mapping`):
            The underlying dictionary.
            If this object is a DictView, its underlying dictionary is used.
        """
        if not isinstance(a_dict, Mapping):
            raise TypeError(
                "The a_dict parameter must be a Mapping, but is: {}".
                format(type(a_dict)))
        if isinstance(a_dict, DictView):
            a_dict = a_dict._dict
        self._dict = a_dict

    @property
    def dict(self):
        """
        The underlying dictionary.

        Access to the underlying dictionary is provided for purposes such as
        conversion to JSON or other cases where the view classes do not work.
        This access should not be used to modify the underlying dictionary.
        """
        return self._dict

    def __repr__(self):
        """
        ``repr(self)``:
        Return a string representation of the view suitable for debugging.

        The underlying dictionary is represented using its ``repr()``
        representation.
        """
        return "{0.__class__.__name__}({1!r})".format(self, self._dict)

    def __getstate__(self):
        """Support for pickling."""
        # Needed on Python 2 due to the use of slots
        a_dict = dict()
        for attr in self.__slots__:
            a_dict[attr] = getattr(self, attr)

        # Support for objects that also have __dict__, e.g. user defined
        # derived classes that did not define __slots__:
        if hasattr(self, '__dict__'):
            for attr in self.__dict__:
                a_dict[attr] = getattr(self, attr)

        return a_dict

    def __setstate__(self, a_dict):
        """Support for unpickling."""
        # Needed on Python 2 due to the use of slots
        for attr in a_dict:
            setattr(self, attr, a_dict[attr])

    def __getitem__(self, key):
        """
        ``self[key]``:
        Return the value of the dictionary item with a key.

        Raises:
          KeyError: An item with the key does not exist.
        """
        return self._dict[key]

    def __len__(self):
        """
        ``len(self)``:
        Return the number of items in the dictionary.

        The return value is the number of items in the underlying dictionary.
        """
        return len(self._dict)

    def __contains__(self, key):
        """
        ``value in self``:
        Return a boolean indicating whether the dictionary contains an item with
        a key.

        The return value indicates whether the underlying dictionary contains an
        item that has the specified key.
        """
        return key in self._dict

    def __reversed__(self):
        """
        ``reversed(self) ...``:
        Return an iterator through the dictionary keys in reversed iteration
        order.

        Added in Python 3.8.

        The returned iterator yields the keys in the underlying dictionary in
        reversed iteration order.
        """
        return reversed(self._dict)

    def get(self, key, default=None):
        """
        Return the value of the dictionary item with a key or a default value.
        """
        return self._dict.get(key, default)

    def has_key(self, key):
        """
        Python 2 only: Return a boolean indicating whether the dictionary
        contains an item with a key.

        Raises:
          AttributeError: The method does not exist on Python 3.
        """
        return self._dict.has_key(key)  # noqa: W601

    # Iteration methods

    def keys(self):
        # pylint: disable=line-too-long
        """
        Return the dictionary keys in iteration order.

        The keys of the underlying dictionary are returned as a view in
        Python 3 and as a list in Python 2.

        See
        `Dictionary View Objects on Python 3 <https://docs.python.org/3/library/stdtypes.html#dictionary-view-objects>`_ for details about view objects.
        """  # noqa: E501
        # pylint: enable=line-too-long
        return self._dict.keys()

    def values(self):
        # pylint: disable=line-too-long
        """
        Return the dictionary values in iteration order.

        The values of the underlying dictionary are returned as a view in
        Python 3 and as a list in Python 2.

        See
        `Dictionary View Objects on Python 3 <https://docs.python.org/3/library/stdtypes.html#dictionary-view-objects>`_ for details about view objects.
        """  # noqa: E501
        # pylint: enable=line-too-long
        return self._dict.values()

    def items(self):
        # pylint: disable=line-too-long
        """
        Return the dictionary items in iteration order.

        Each returned item is a tuple of key and value.
        The items of the underlying dictionary are returned as a view in
        Python 3 and as a list in Python 2.

        See
        `Dictionary View Objects on Python 3 <https://docs.python.org/3/library/stdtypes.html#dictionary-view-objects>`_ for details about view objects.
        """  # noqa: E501
        # pylint: enable=line-too-long
        return self._dict.items()

    def iterkeys(self):
        """
        Python 2 only: Return an iterator through the dictionary keys in
        iteration order.

        Raises:
          AttributeError: The method does not exist on Python 3.
        """
        return self._dict.iterkeys()

    def itervalues(self):
        """
        Python 2 only: Return an iterator through the dictionary values in
        iteration order.

        Raises:
          AttributeError: The method does not exist on Python 3.
        """
        return self._dict.itervalues()

    def iteritems(self):
        """
        Python 2 only: Return an iterator through the dictionary items in
        iteration order.

        Each item is a tuple of key and value.

        Raises:
          AttributeError: The method does not exist on Python 3.
        """
        return self._dict.iteritems()

    def viewkeys(self):
        # pylint: disable=line-too-long
        """
        Python 2 only: Return a view on the dictionary keys in iteration order.

        The keys of the underlying dictionary are returned as a view.

        See
        `Dictionary View Objects on Python 2 <https://docs.python.org/2/library/stdtypes.html#dictionary-view-objects>`_ for details about view objects.

        Raises:
          AttributeError: The method does not exist on Python 3.
        """  # noqa: E501
        # pylint: enable=line-too-long
        return self._dict.viewkeys()

    def viewvalues(self):
        # pylint: disable=line-too-long
        """
        Python 2 only: Return a view on the dictionary values in iteration
        order.

        The values of the underlying dictionary are returned as a view.

        See
        `Dictionary View Objects on Python 2 <https://docs.python.org/2/library/stdtypes.html#dictionary-view-objects>`_ for details about view objects.

        Raises:
          AttributeError: The method does not exist on Python 3.
        """  # noqa: E501
        # pylint: enable=line-too-long
        return self._dict.viewvalues()

    def viewitems(self):
        # pylint: disable=line-too-long
        """
        Python 2 only: Return a view on the dictionary items in iteration order.

        Each returned item is a tuple of key and value.
        The items of the underlying dictionary are returned as a view.

        See
        `Dictionary View Objects on Python 2 <https://docs.python.org/2/library/stdtypes.html#dictionary-view-objects>`_ for details about view objects.

        Raises:
          AttributeError: The method does not exist on Python 3.
        """  # noqa: E501
        # pylint: enable=line-too-long
        return self._dict.viewitems()

    def __iter__(self):
        """
        Return an iterator through the dictionary keys in iteration order.
        """
        return iter(self._dict)

    def copy(self):
        """
        Return a new view on a shallow copy of the dictionary.

        The returned :class:`DictView` object is a new view object on a
        dictionary object of the type of the underlying dictionary.

        If the dictionary type is immutable, the returned dictionary object may
        be the underlying dictionary object. If the dictionary type is mutable,
        the returned dictionary is a new dictionary object that is a shallow
        copy of the underlying dictionary object.
        """
        org_class = self._dict.__class__
        new_dict = org_class(self._dict)  # May be same object if immutable
        return DictView(new_dict)

    def __eq__(self, other):
        """
        ``self == other``:
        Return a boolean indicating whether the dictionary is equal to the
        other object.

        Compared are the underlying dictionary of the left hand view object
        and the right hand object (or in case of a DictView, its underlying
        dictionary).
        """
        # pylint: disable=protected-access
        other_dict = other._dict if isinstance(other, DictView) else other
        return self._dict == other_dict

    def __ne__(self, other):
        """
        ``self != other``:
        Return a boolean indicating whether the dictionary is not equal to the
        other dictionary.

        Compared are the underlying dictionary of the left hand view object
        and the right hand object (or in case of a DictView, its underlying
        dictionary).
        """
        # pylint: disable=protected-access
        other_dict = other._dict if isinstance(other, DictView) else other
        return self._dict != other_dict

    def __gt__(self, other):
        """
        ``self > other``:
        Return a boolean indicating whether the dictionary is greater than
        the other dictionary.

        Whether ordering comparison is supported and how ordering is defined
        depends on the underlying dictionary and the other dictionary. For
        example, the standard Python :class:`dict` class does not support
        ordering comparisons on Python 3.

        The other object must be a :class:`dict` or :class:`DictView`.

        Raises:
          TypeError: The other object is not a dict or DictView.
          TypeError: The underlying dictionary does not support ordering
            comparisons.
        """
        # pylint: disable=protected-access
        other_dict = other._dict if isinstance(other, DictView) else other
        return self._dict > other_dict

    def __lt__(self, other):
        """
        ``self < other``:
        Return a boolean indicating whether the dictionary is less than
        the other dictionary.

        Whether ordering comparison is supported and how ordering is defined
        depends on the underlying dictionary and the other dictionary. For
        example, the standard Python :class:`dict` class does not support
        ordering comparisons on Python 3.

        The other object must be a :class:`dict` or :class:`DictView`.

        Raises:
          TypeError: The other object is dict a set or DictView.
          TypeError: The underlying dictionary does not support ordering
            comparisons.
        """
        # pylint: disable=protected-access
        other_dict = other._dict if isinstance(other, DictView) else other
        return self._dict < other_dict

    def __ge__(self, other):
        """
        ``self >= other``:
        Return a boolean indicating whether the dictionary is greater than or
        equal to the other dictionary.

        Whether ordering comparison is supported and how ordering is defined
        depends on the underlying dictionary and the other dictionary. For
        example, the standard Python :class:`dict` class does not support
        ordering comparisons on Python 3.

        The other object must be a :class:`dict` or :class:`DictView`.

        Raises:
          TypeError: The other object is not a dict or DictView.
          TypeError: The underlying dictionary does not support ordering
            comparisons.
        """
        # pylint: disable=protected-access
        other_dict = other._dict if isinstance(other, DictView) else other
        return self._dict >= other_dict

    def __le__(self, other):
        """
        ``self <= other``:
        Return a boolean indicating whether the dictionary is less than or
        equal to the other dictionary.

        Whether ordering comparison is supported and how ordering is defined
        depends on the underlying dictionary and the other dictionary. For
        example, the standard Python :class:`dict` class does not support
        ordering comparisons on Python 3.

        The other object must be a :class:`dict` or :class:`DictView`.

        Raises:
          TypeError: The other object is not a dict or DictView.
          TypeError: The underlying dictionary does not support ordering
            comparisons.
        """
        # pylint: disable=protected-access
        other_dict = other._dict if isinstance(other, DictView) else other
        return self._dict <= other_dict

    def __hash__(self):
        """
        ``hash(self)``:
        Return a hash value for the dictionary.

        Whether hashing is supported depends on the underlying dictionary. For
        example, the standard Python :class:`dict` class does not support
        hashing.

        Raises:
          TypeError: The underlying dictionary does not support hashing.
        """
        return hash(self._dict)

    def __or__(self, other):
        """
        ``self | other``:
        Return a new view on the merged dictionary and other dictionary.

        Added in Python 3.9.

        The returned :class:`DictView` object is a view on a new dictionary
        object of the type of the left hand operand that contains all the items
        from the underlying dictionary of the left hand operand, updated by the
        items from the other dictionary (or in case of a DictView, its
        underlying dictionary).

        The other object must be a :class:`dict` or :class:`DictView`.

        The dictionary and the other dictionary are not changed.

        Raises:
          TypeError: The other object is not a dict or DictView.
        """
        # pylint: disable=protected-access
        other_dict = other._dict if isinstance(other, DictView) else other
        new_dict = self._dict | other_dict
        return DictView(new_dict)

    def __ror__(self, other):
        """
        ``other | self``:
        Return a new view on the merged dictionary and other dictionary.

        Added in Python 3.9.

        This method is a fallback and is called only if the left operand does
        not support the operation.

        The returned :class:`DictView` object is a view on a new dictionary
        object of the type of the right hand operand that contains all the items
        from the underlying dictionary of the right hand operand, updated by the
        items from the other dictionary (or in case of a DictView, its
        underlying dictionary).

        The other object must be a :class:`dict` or :class:`DictView`.

        The dictionary and the other dictionary are not changed.

        Raises:
          TypeError: The other object is not a dict or DictView.
        """
        return self.__or__(other)


# Remove methods that should be present only under certain conditions, and when
# building the documentation.

if not _DICT_SUPPORTS_ITER_VIEW and not _BUILDING_DOCS:
    del DictView.iterkeys
    del DictView.itervalues
    del DictView.iteritems
    del DictView.viewkeys
    del DictView.viewvalues
    del DictView.viewitems

if not _DICT_SUPPORTS_HAS_KEY and not _BUILDING_DOCS:
    del DictView.has_key

if not _DICT_SUPPORTS_REVERSED and not _BUILDING_DOCS:
    del DictView.__reversed__

if not _DICT_SUPPORTS_OR and not _BUILDING_DOCS:
    del DictView.__or__
    del DictView.__ror__



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
An immutable set view.
"""

from __future__ import print_function, absolute_import

try:
    from collections.abc import Set
except ImportError:
    # Python 2
    from collections import Set  # pylint: disable=deprecated-class

__all__ = ['SetView']


class SetView(Set):
    # pylint: disable=line-too-long
    """
    An immutable set view.

    Derived from :class:`~py3:collections.abc.Set`.

    This class provides an immutable view on a possibly mutable set object.
    The set object must be an instance of :class:`~py3:collections.abc.Set`,
    e.g. :class:`set`, or a user-defined class.

    This can be used for example when a class maintains a set that should be
    made available to users of the class without allowing them to modify the
    set.

    In the description of this class, the term 'view' always refers to the
    :class:`SetView` object, and the term 'set' or 'underlying set' refers to
    the set object the view is based on.

    The :class:`SetView` class supports the complete behavior of Python class
    :class:`set`, except for any methods that would modify the set.
    Note that the non-modifying methods of class :class:`set` are a superset of
    the methods defined for the abstract class :class:`~py3:collections.abc.Set`
    (the methods are listed in the table at the top of the linked page).

    The view is "live": Since the view class delegates all operations to the
    underlying set, any modification of the underlying set object will be
    visible in the view object.

    Note that only the view object is immutable, not necessarily its items.
    So if the items in the underlying set are mutable objects, they can be
    modified through the view.

    Note that in Python, augmented assignment (e.g. ``x += y``) is not
    guaranteed to modify the left hand object in place, but can result in the
    left hand name being bound to a new object (like in ``x = x + y``).
    For details, see
    `object.__iadd__() <https://docs.python.org/3/reference/datamodel.html#object.__iadd__>`_.

    For the SetView class, augmented assignment is supported and results in
    binding the left hand name to a new SetView object.
    """  # noqa: E501
    # pylint: enable=line-too-long

    __slots__ = ['_set']

    def __init__(self, a_set):
        """
        Parameters:

          a_set (:class:`~py3:collections.abc.Set`):
            The underlying set.
            If this object is a SetView, its underlying set is used.
        """
        if not isinstance(a_set, Set):
            raise TypeError(
                "The a_set parameter must be a Set, but is: {}".
                format(type(a_set)))
        if isinstance(a_set, SetView):
            a_set = a_set._set
        self._set = a_set

    @property
    def set(self):
        """
        The underlying set.

        Access to the underlying set is provided for purposes such as
        conversion to JSON or other cases where the view classes do not work.
        This access should not be used to modify the underlying set.
        """
        return self._set

    def __repr__(self):
        """
        ``repr(self)``:
        Return a string representation of the view suitable for debugging.

        The underlying set is represented using its ``repr()`` representation.
        """
        return "{0.__class__.__name__}({1!r})".format(self, self._set)

    def __getstate__(self):
        """Support for pickling."""
        # Needed on Python 2 due to the use of slots
        a_dict = dict()
        for attr in self.__slots__:
            a_dict[attr] = getattr(self, attr)

        # Support for objects that also have __dict__, e.g. user defined
        # derived classes that did not define __slots__:
        if hasattr(self, '__dict__'):
            for attr in self.__dict__:
                a_dict[attr] = getattr(self, attr)

        return a_dict

    def __setstate__(self, a_dict):
        """Support for unpickling."""
        # Needed on Python 2 due to the use of slots
        for attr in a_dict:
            setattr(self, attr, a_dict[attr])

    def __len__(self):
        """
        ``len(self)``:
        Return the number of items in the set.

        The return value is the number of items in the underlying set.
        """
        return len(self._set)

    def __contains__(self, value):
        """
        ``value in self``:
        Return a boolean indicating whether the set contains a value.

        The return value indicates whether the underlying set contains an item
        that is equal to the value.
        """
        return value in self._set

    def __iter__(self):
        """
        ``iter(self) ...``:
        Return an iterator through the set.

        The returned iterator yields the items in the underlying set in its
        iteration order.
        """
        return iter(self._set)

    def __and__(self, other):
        """
        ``self & other``:
        Return a new view on the intersection of the set and the other set.

        The returned :class:`SetView` object is a view on a new set object of
        the type of the left hand operand that contains the items that are in
        the underlying set of the left hand operand and in the other set
        (or in case of a SetView, its underlying set).

        The other object must be a :class:`set` or :class:`SetView`.

        The set and the other set are not changed.

        Raises:
          TypeError: The other object is not a set or SetView.
        """
        # pylint: disable=protected-access
        other_set = other._set if isinstance(other, SetView) else other
        new_set = self._set & other_set
        return SetView(new_set)

    def __rand__(self, other):
        """
        ``other & self``:
        Return a new view on the intersection of the set and the other set.

        This method is a fallback and is called only if the left operand does
        not support the operation.

        The returned :class:`SetView` object is a view on a new set object of
        the type of the right hand operand that contains the items that are in
        the underlying set of the right hand operand and in the other set
        (or in case of a SetView, its underlying set).

        The other object must be a :class:`set` or :class:`SetView`.

        The set and the other set are not changed.

        Raises:
          TypeError: The other object is not a set or SetView.
        """
        return self.__and__(other)

    def intersection(self, *others):
        """
        Return a new view on the intersection of the set and the other
        iterables.

        The returned :class:`SetView` object is a view on a new set object of
        the type of the underlying set that contains the items that are in the
        underlying set and in the other iterables (or in case of SetView
        objects, their underlying sets).

        The other objects must be :term:`iterables <py3:iterable>`.

        The set and the other iterables are not changed.

        Raises:
          TypeError: The other objects are not all iterables.
        """
        # pylint: disable=protected-access
        other_sets = [other._set if isinstance(other, SetView) else other
                      for other in others]
        new_set = self._set.intersection(*other_sets)
        return SetView(new_set)

    def __or__(self, other):
        """
        ``self | other``:
        Return a new view on the union of the set and the other set.

        The returned :class:`SetView` object is a view on a new set object of
        the type of the left hand operand that contains all the (unique) items
        from the underlying set of the left hand operand and the other set
        (or in case of a SetView, its underlying set).

        The other object must be a :class:`set` or :class:`SetView`.

        The set and the other set are not changed.

        Raises:
          TypeError: The other object is not a set or SetView.
        """
        # pylint: disable=protected-access
        other_set = other._set if isinstance(other, SetView) else other
        new_set = self._set | other_set
        return SetView(new_set)

    def __ror__(self, other):
        """
        ``other | self``:
        Return a new view on the union of the set and the other set.

        This method is a fallback and is called only if the left operand does
        not support the operation.

        The returned :class:`SetView` object is a view on a new set object of
        the type of the right hand operand that contains all the (unique) items
        from the underlying set of the right hand operand and the other set
        (or in case of a SetView, its underlying set).

        The other object must be a :class:`set` or :class:`SetView`.

        The set and the other set are not changed.

        Raises:
          TypeError: The other object is not a set or SetView.
        """
        return self.__or__(other)

    def union(self, *others):
        """
        Return a new view on the union of the set and the other iterables.

        The returned :class:`SetView` object is a view on a new set object of
        the type of the underlying set that contains all the (unique) items from
        the underlying set and the other iterables (or in case of SetView
        objects, their underlying sets).

        The other objects must be :term:`iterables <py3:iterable>`.

        The set and the other iterables are not changed.

        Raises:
          TypeError: The other objects are not all iterables.
        """
        # pylint: disable=protected-access
        other_sets = [other._set if isinstance(other, SetView) else other
                      for other in others]
        new_set = self._set.union(*other_sets)
        return SetView(new_set)

    def __sub__(self, other):
        """
        ``self - other``:
        Return a new view on the difference of the set and the other set.

        The returned :class:`SetView` object is a view on a new set object of
        the type of the left hand operand that contains the items that are in
        the underlying set of the left hand operand but not in the other set
        (or in case of a SetView, its underlying set).

        The other object must be a :class:`set` or :class:`SetView`.

        The set and the other set are not changed.

        Raises:
          TypeError: The other object is not a set or SetView.
        """
        # pylint: disable=protected-access
        other_set = other._set if isinstance(other, SetView) else other
        new_set = self._set - other_set
        return SetView(new_set)

    def __rsub__(self, other):
        """
        ``other - self``:
        Return a new view on the difference of the other set and the set.

        This method is a fallback and is called only if the left operand does
        not support the operation.

        The returned :class:`SetView` object is a view on a new set object of
        the type of the left hand operand that contains the items that are in
        the other set (or in case of a SetView, its underlying set) but not in
        the underlying set of the left hand operand.

        The other object must be a :class:`set` or :class:`SetView`.

        The set and the other set are not changed.

        Raises:
          TypeError: The other object is not a set or SetView.
        """
        # pylint: disable=protected-access
        other_set = other._set if isinstance(other, SetView) else other
        new_set = set()
        for item in other_set:
            if item not in self._set:
                new_set.add(item)
        return SetView(new_set)

    def difference(self, *others):
        """
        Return a new view on the difference of the set and the other
        iterables.

        The returned :class:`SetView` object is a view on a new set object of
        the type of the underlying set that contains the items that are in the
        underlying set but not in any of the other iterables (or in case of
        SetView objects, their underlying sets).

        The other objects must be :term:`iterables <py3:iterable>`.

        The set and the other iterables are not changed.

        Raises:
          TypeError: The other objects are not all iterables.
        """
        # pylint: disable=protected-access
        other_sets = [other._set if isinstance(other, SetView) else other
                      for other in others]
        new_set = self._set.difference(*other_sets)
        return SetView(new_set)

    def __xor__(self, other):
        """
        ``self ^ other``:
        Return a new view on the symmetric difference of the set and the other
        set.

        The returned :class:`SetView` object is a view on a new set object of
        the type of the left hand operand that contains the items that are in
        either the underlying set of the left hand operand or in the other set
        (or in case of a SetView, its underlying set), but not in both.

        The other object must be a :class:`set` or :class:`SetView`.

        The set and the other set are not changed.

        Raises:
          TypeError: The other object is not a set or SetView.
        """
        # pylint: disable=protected-access
        other_set = other._set if isinstance(other, SetView) else other
        new_set = self._set ^ other_set
        return SetView(new_set)

    def __rxor__(self, other):
        """
        ``other ^ self``:
        Return a new view on the symmetric difference of the set and the other
        set.

        This method is a fallback and is called only if the left operand does
        not support the operation.

        The returned :class:`SetView` object is a view on a new set object of
        the type of the right hand operand that contains the items that are in
        either the underlying set of the right hand operand or in the other set
        (or in case of a SetView, its underlying set), but not in both.

        The other object must be a :class:`set` or :class:`SetView`.

        The set and the other set are not changed.

        Raises:
          TypeError: The other object is not a set or SetView.
        """
        return self.__xor__(other)

    def symmetric_difference(self, other):
        """
        Return a new view on the symmetric difference of the set and the other
        iterable.

        The returned :class:`SetView` object is a view on a new set object of
        the type of the underlying set that contains the items that are in
        either the underlying set or in the other iterable (or in case of a
        SetView, its underlying set), but not in both.

        The other object must be an :term:`py3:iterable`.

        The set and the other iterable are not changed.

        Raises:
          TypeError: The other object is not an iterable.
        """
        # pylint: disable=protected-access
        other_set = other._set if isinstance(other, SetView) else other
        new_set = self._set.symmetric_difference(other_set)
        return SetView(new_set)

    def __eq__(self, other):
        """
        ``self == other``:
        Return a boolean indicating whether the set is equal to the other set.

        The return value indicates whether the items in the underlying set are
        equal to the items in the other set (or in case of a SetView, its
        underlying set).

        The other object must be a :class:`set` or :class:`SetView`.

        Raises:
          TypeError: The other object is not a set or SetView.
        """
        # pylint: disable=protected-access
        other_set = other._set if isinstance(other, SetView) else other
        return self._set == other_set

    def __ne__(self, other):
        """
        ``self != other``:
        Return a boolean indicating whether the set is not equal to the other
        set.

        The return value indicates whether the items in the underlying set are
        not equal to the items in the other set (or in case of a SetView, its
        underlying set).

        The other object must be a :class:`set` or :class:`SetView`.

        Raises:
          TypeError: The other object is not a set or SetView.
        """
        # pylint: disable=protected-access
        other_set = other._set if isinstance(other, SetView) else other
        return self._set != other_set

    def __gt__(self, other):
        """
        ``self > other``:
        Return a boolean indicating whether the set is a proper superset of the
        other set.

        The return value indicates whether the underlying set is a proper
        superset of the other set (or in case of a SetView, its underlying set).

        The other object must be a :class:`set` or :class:`SetView`.

        Raises:
          TypeError: The other object is not a set or SetView.
        """
        # pylint: disable=protected-access
        other_set = other._set if isinstance(other, SetView) else other
        return self._set > other_set

    def __lt__(self, other):
        """
        ``self < other``:
        Return a boolean indicating whether the set is a proper subset of the
        other set.

        The return value indicates whether the underlying set is a proper
        subset of the other set (or in case of a SetView, its underlying set).

        The other object must be a :class:`set` or :class:`SetView`.

        Raises:
          TypeError: The other object is not a set or SetView.
        """
        # pylint: disable=protected-access
        other_set = other._set if isinstance(other, SetView) else other
        return self._set < other_set

    def __ge__(self, other):
        """
        ``self >= other``:
        Return a boolean indicating whether the set is an inclusive superset
        of the other set.

        The return value indicates whether every item in the other set
        (or in case of a SetView, its underlying set) is in the underlying set.

        The other object must be a :class:`set` or :class:`SetView`.

        Raises:
          TypeError: The other object is not a set or SetView.
        """
        # pylint: disable=protected-access
        other_set = other._set if isinstance(other, SetView) else other
        return self._set >= other_set

    def issuperset(self, other):
        """
        Return a boolean indicating whether the set is an inclusive superset
        of the other iterable.

        The return value indicates whether every item in the other iterable
        (or in case of a SetView, its underlying set) is in the underlying set.

        The other object must be an :term:`py3:iterable`.

        Raises:
          TypeError: The other object is not an iterable.
        """
        # pylint: disable=protected-access
        other_set = other._set if isinstance(other, SetView) else other
        return self._set.issuperset(other_set)

    def __le__(self, other):
        """
        ``self <= other``:
        Return a boolean indicating whether the set is an inclusive subset
        of the other set.

        The return value indicates whether every item in the underlying
        set is in the other set (or in case of a SetView, its underlying set).

        The other object must be a :class:`set` or :class:`SetView`.

        Raises:
          TypeError: The other object is not a set or SetView.
        """
        # pylint: disable=protected-access
        other_set = other._set if isinstance(other, SetView) else other
        return self._set <= other_set

    def issubset(self, other):
        """
        Return a boolean indicating whether the set is an inclusive subset
        of the other iterable.

        The return value indicates whether every item in the underlying set
        is in the other iterable (or in case of a SetView, its underlying set).

        The other object must be an :term:`py3:iterable`.

        Raises:
          TypeError: The other object is not an iterable.
        """
        # pylint: disable=protected-access
        other_set = other._set if isinstance(other, SetView) else other
        return self._set.issubset(other_set)

    def isdisjoint(self, other):
        """
        Return a boolean indicating whether the set does not intersect with
        the other iterable.

        The return value indicates whether the underlying set has no items in
        common with the other iterable (or in case of a SetView, its underlying
        set).

        The other object must be an :term:`py3:iterable`.

        Raises:
          TypeError: The other object is not an iterable.
        """
        # pylint: disable=protected-access
        other_set = other._set if isinstance(other, SetView) else other
        return self._set.isdisjoint(other_set)

    def copy(self):
        """
        Return a new view on a shallow copy of the set.

        The returned :class:`SetView` object is a new view object on a set
        object of the type of the underlying set.

        If the set type is immutable, the returned set object may be the
        underlying set object. If the set type is mutable, the returned set is
        a new set object that is a shallow copy of the underlying set object.
        """
        org_class = self._set.__class__
        new_set = org_class(self._set)  # May be same object if immutable
        return SetView(new_set)

    def __hash__(self):
        """
        ``hash(self)``:
        Return a hash value for the set.

        Whether hashing is supported depends on the underlying set. For
        example, the standard Python :class:`set` class does not support
        hashing, but the standard Python :class:`frozenset` class does.

        Raises:
          TypeError: The underlying set does not support hashing.
        """
        return hash(self._set)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test the DictView class.
"""

from __future__ import absolute_import

import sys
import re
import pickle
from collections import OrderedDict
try:
    from collections.abc import KeysView, ValuesView, ItemsView, Iterator, \
        MutableMapping, Mapping
except ImportError:
    # pylint: disable=deprecated-class
    from collections import KeysView, ValuesView, ItemsView, Iterator, \
        MutableMapping, Mapping
import pytest
from nocasedict import NocaseDict, HashableMixin

from ..utils.simplified_test_function import simplified_test_function

# pylint: disable=wrong-import-position, wrong-import-order, invalid-name
from ..utils.import_installed import import_installed
immutable_views = import_installed('immutable_views')
from immutable_views import DictView  # noqa: E402
# pylint: enable=wrong-import-position, wrong-import-order, invalid-name


class HashableDict(HashableMixin, NocaseDict):
    """Hashable dictionary, for testing"""
    pass


class NorDict(Mapping):
    """Dictionary without OR operator, for testing"""

    def __init__(self, a_dict=None):
        if a_dict is None:
            a_dict = {}
        self._dict = dict(a_dict)

    def __getitem__(self, key):
        return self._dict[key]

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)


class OrDict(NorDict):
    """Dictionary with OR operator, for testing"""

    def __or__(self, other):
        return self._dict | dict(other)

    def __ror__(self, other):
        return dict(other) | self._dict


class DerivedDict(dict):
    """Dictionary derived from dict with additional attribute"""

    def __init__(self, *args, **kwargs):
        super(DerivedDict, self).__init__(*args, **kwargs)
        self._dummy = 'dummy'

    def __eq__(self, other):
        super_equal = super(DerivedDict, self).__eq__(other)
        if isinstance(other, DerivedDict):
            return super_equal and self._dummy == other._dummy
        return super_equal


class DerivedDictView(DictView):
    """Dictionary view derived from DictView with additional attribute"""

    def __init__(self, *args, **kwargs):
        super(DerivedDictView, self).__init__(*args, **kwargs)
        self._dummy = 'dummy'

    def __eq__(self, other):
        super_equal = super(DerivedDictView, self).__eq__(other)
        if isinstance(other, DerivedDictView):
            return super_equal and self._dummy == other._dummy
        return super_equal


PY2 = sys.version_info[0] == 2

# Indicates Python dict supports lt/gt comparison (between dicts)
DICT_SUPPORTS_COMPARISON = sys.version_info[0:2] == (2, 7)

# Indicates Python dict is ordered
DICT_IS_ORDERED = sys.version_info[0:2] >= (3, 7)

# Indicates Python dict issues UserWarning about not preserving order
# of items in kwargs or in standard dict
DICT_WARNS_ORDER = sys.version_info[0:2] <= (3, 6)

# Indicates Python dict supports the iter..() and view..() methods
DICT_SUPPORTS_ITER_VIEW = sys.version_info[0:2] == (2, 7)

# Indicates Python dict supports the has_key() method
DICT_SUPPORTS_HAS_KEY = sys.version_info[0:2] == (2, 7)

# Indicates Python dict supports the __reversed__() method
DICT_SUPPORTS_REVERSED = sys.version_info[0:2] >= (3, 8)

# Indicates Python dict supports the __or__/__ror__() methods
DICT_SUPPORTS_OR = sys.version_info[0:2] >= (3, 9)

# Used as indicator not to pass an argument in the testcases.
# Note this has nothing to do with the _OMITTED flag in _immutable_views.py and
# could be a different value.
_OMIT_ARG = object()


class NonEquatable(object):
    # pylint: disable=too-few-public-methods
    """
    Class that raises TypeError when comparing its objects for equality.
    """

    def __eq__(self, other):
        raise TypeError("Cannot compare %s to %s" % (type(self), type(other)))

    def __ne__(self, other):
        raise TypeError("Cannot compare %s to %s" % (type(self), type(other)))


TESTCASES_DICTVIEW_INIT = [

    # Testcases for DictView.__init__() / ncd=DictView()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * init_args: Tuple of positional arguments to DictView().
    #   * init_kwargs: Dict of keyword arguments to DictView().
    #   * exp_dict: Expected resulting dictionary, as OrderedDict.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty underlying dictionaries
    (
        "From empty dict as positional arg",
        dict(
            init_args=({},),
            init_kwargs={},
            exp_dict=DictView({}),
            verify_order=True,
        ),
        None, None, True
    ),
    (
        "From empty OrderedDict as positional arg",
        dict(
            init_args=(OrderedDict(),),
            init_kwargs={},
            exp_dict=DictView(OrderedDict()),
            verify_order=True,
        ),
        None, None, True
    ),
    (
        "From DictView of empty dict as positional arg",
        dict(
            init_args=(DictView({}),),
            init_kwargs={},
            exp_dict=DictView({}),
            verify_order=True,
        ),
        None, None, True
    ),

    # Non-empty underlying dictionaries
    (
        "From non-empty dict as positional arg",
        dict(
            init_args=({'Dog': 'Cat', 'Budgie': 'Fish'},),
            init_kwargs={},
            exp_dict=DictView({'Dog': 'Cat', 'Budgie': 'Fish'}),
            verify_order=False,
        ),
        None, None, True
    ),
    (
        "From DictView of non-empty dict as positional arg",
        dict(
            init_args=(DictView({'Dog': 'Cat', 'Budgie': 'Fish'}),),
            init_kwargs={},
            exp_dict=DictView({'Dog': 'Cat', 'Budgie': 'Fish'}),
            verify_order=False,
        ),
        None, None, True
    ),

    # Error cases with inputs that are invalid for the view but would be valid
    # for a real dictionary
    (
        "From no args (a_dict parameter required)",
        dict(
            init_args=(),
            init_kwargs={},
            exp_dict=None,
            verify_order=None,
        ),
        TypeError, None, True
    ),
    (
        "From None as positional arg (Must be Mapping type)",
        dict(
            init_args=(None,),
            init_kwargs={},
            exp_dict=None,
            verify_order=None,
        ),
        TypeError, None, True
    ),
    (
        "From empty list as positional arg (Must be Mapping type)",
        dict(
            init_args=(list(),),
            init_kwargs={},
            exp_dict=None,
            verify_order=None,
        ),
        TypeError, None, True
    ),
    (
        "From empty tuple as positional arg (Must be Mapping type)",
        dict(
            init_args=(tuple(),),
            init_kwargs={},
            exp_dict=None,
            verify_order=None,
        ),
        TypeError, None, True
    ),
    (
        "From list as positional arg (Must be Mapping type)",
        dict(
            init_args=([('Dog', 'Cat'), ('Budgie', 'Fish')],),
            init_kwargs={},
            exp_dict=None,
            verify_order=None,
        ),
        TypeError, None, True
    ),
    (
        "From tuple as positional arg (Must be Mapping type)",
        dict(
            init_args=((('Dog', 'Cat'), ('Budgie', 'Fish')),),
            init_kwargs={},
            exp_dict=None,
            verify_order=None,
        ),
        TypeError, None, True
    ),
    (
        "From keyword args (Must be Mapping type)",
        dict(
            init_args=(),
            init_kwargs={'Dog': 'Cat', 'Budgie': 'Fish'},
            exp_dict=None,
            verify_order=None,
        ),
        TypeError, None, True
    ),
    (
        "From list as positional arg and keyword args (Must be Mapping type)",
        dict(
            init_args=([('Dog', 'Cat')],),
            init_kwargs={'Budgie': 'Fish'},
            exp_dict=None,
            verify_order=None,
        ),
        TypeError, None, True
    ),
    (
        "From tuple as positional arg and keyword args (Must be Mapping type)",
        dict(
            init_args=((('Dog', 'Cat'),),),
            init_kwargs={'Budgie': 'Fish'},
            exp_dict=None,
            verify_order=None,
        ),
        TypeError, None, True
    ),
    (
        "From dict as positional arg and keyword args (Must be Mapping type)",
        dict(
            init_args=({'Dog': 'Cat'},),
            init_kwargs={'Budgie': 'Fish'},
            exp_dict=None,
            verify_order=None,
        ),
        TypeError, None, True
    ),

    # Other error cases
    (
        "String as positional arg (Must be Mapping type)",
        dict(
            init_args=('illegal',),
            init_kwargs={},
            exp_dict=None,
            verify_order=None,
        ),
        TypeError, None, True
    ),
    (
        "Integer as positional arg (Must be Mapping type)",
        dict(
            init_args=(42,),
            init_kwargs={},
            exp_dict=None,
            verify_order=None,
        ),
        TypeError, None, True
    ),
    (
        "Two positional args (too many args)",
        dict(
            init_args=(list(), list()),
            init_kwargs={},
            exp_dict=None,
            verify_order=None,
        ),
        TypeError, None, True
    ),
    (
        "List as positional arg, whose item has only one item "
        "(Must be Mapping type)",
        dict(
            init_args=([('Dog',)],),
            init_kwargs={},
            exp_dict=None,
            verify_order=None,
        ),
        TypeError, None, True
    ),
    (
        "List as positional arg, whose item has too many items "
        "(Must be Mapping type)",
        dict(
            init_args=([('Dog', 'Cat', 'bad')],),
            init_kwargs={},
            exp_dict=None,
            verify_order=None,
        ),
        TypeError, None, True
    ),
    (
        "Tuple as positional arg, whose item has only one item "
        "(Must be Mapping type)",
        dict(
            init_args=((('Dog',),),),
            init_kwargs={},
            exp_dict=None,
            verify_order=None,
        ),
        TypeError, None, True
    ),
    (
        "Tuple as positional arg, whose item has too many items "
        "(Must be Mapping type)",
        dict(
            init_args=((('Dog', 'Cat', 'bad'),),),
            init_kwargs={},
            exp_dict=None,
            verify_order=None,
        ),
        TypeError, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_INIT)
@simplified_test_function
def test_DictView_init(
        testcase, init_args, init_kwargs, exp_dict, verify_order):
    """
    Test function for DictView.__init__() / ncd=DictView()
    """

    # The code to be tested
    act_dict = DictView(*init_args, **init_kwargs)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    # The verification below also uses some DictView features, but that is
    # unavoidable if we want to work through the public interface:

    act_items = []
    for key in act_dict:  # Uses DictView iteration
        act_value = act_dict[key]  # Uses DictView getitem
        assert key in exp_dict, "Unexpected extra key %r" % key
        exp_value = exp_dict[key]
        assert act_value == exp_value, "Unexpected value at key %r" % key
        act_items.append((key, act_value))

    exp_items = []
    for key in exp_dict:
        exp_value = exp_dict[key]
        # Next line uses DictView contains:
        assert key in act_dict, "Unexpected missing key %r" % key
        act_value = act_dict[key]  # Uses DictView getitem
        assert act_value == exp_value, "Unexpected value at key %r" % key
        exp_items.append((key, exp_value))

    if verify_order:
        assert act_items == exp_items, "Unexpected order of items"


TESTCASES_DICTVIEW_GETITEM = [

    # Testcases for DictView.__getitem__() / ncd[key]

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj: DictView object to be used for the test.
    #   * key: Key to be used for the test.
    #   * exp_value: Expected value for the key.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty underlying dictionary
    (
        "Empty underlying dict, with None key (not found)",
        dict(
            obj=DictView({}),
            key=None,
            exp_value=None,
        ),
        KeyError, None, True
    ),
    (
        "Empty underlying dict, with integer key (not found)",
        dict(
            obj=DictView({}),
            key=1234,
            exp_value=None,
        ),
        KeyError, None, True
    ),
    (
        "Empty underlying dict, with empty string key (not found)",
        dict(
            obj=DictView({}),
            key='',
            exp_value=None,
        ),
        KeyError, None, True
    ),
    (
        "Empty underlying dict, with non-empty key (not found)",
        dict(
            obj=DictView({}),
            key='Dog',
            exp_value=None,
        ),
        KeyError, None, True
    ),

    # Non-empty underlying dictionary
    (
        "Non-empty underlying dict, with None key (not found)",
        dict(
            obj=DictView(OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')])),
            key=None,
            exp_value=None,
        ),
        KeyError, None, True
    ),
    (
        "Non-empty underlying dict, with empty string key (not found)",
        dict(
            obj=DictView(OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')])),
            key='',
            exp_value=None,
        ),
        KeyError, None, True
    ),
    (
        "Non-empty underlying dict, with non-existing key (not found)",
        dict(
            obj=DictView(OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')])),
            key='invalid',
            exp_value=None,
        ),
        KeyError, None, True
    ),
    (
        "Non-empty underlying dict, with existing key",
        dict(
            obj=DictView(OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')])),
            key='Dog',
            exp_value='Cat',
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_GETITEM)
@simplified_test_function
def test_DictView_getitem(testcase, obj, key, exp_value):
    """
    Test function for DictView.__getitem__() / ncd[key]
    """

    # The code to be tested
    act_value = obj[key]

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_value == exp_value, "Unexpected value at key %r" % key


TESTCASES_DICTVIEW_LEN = [

    # Testcases for DictView.__len__() / len(ncd)

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj: DictView object to be used for the test.
    #   * exp_len: Expected len() value.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty dict",
        dict(
            obj=DictView({}),
            exp_len=0,
        ),
        None, None, True
    ),
    (
        "Dict with two items",
        dict(
            obj=DictView(OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')])),
            exp_len=2,
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_LEN)
@simplified_test_function
def test_DictView_len(testcase, obj, exp_len):
    """
    Test function for DictView.__len__() / len(ncd)
    """

    # The code to be tested
    act_len = len(obj)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_len == exp_len


TESTCASES_DICTVIEW_CONTAINS = [

    # Testcases for DictView.__contains__() / key in ncd

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj: DictView object to be used for the test.
    #   * key: Key to be used for the test.
    #   * exp_result: Expected result (bool).
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty DictView
    (
        "Empty dict, with None key",
        dict(
            obj=DictView({}),
            key=None,
            exp_result=False,
        ),
        None, None, True
    ),
    (
        "Empty dict, with integer key",
        dict(
            obj=DictView({}),
            key=1234,
            exp_result=False,
        ),
        None, None, True
    ),
    (
        "Empty dict, with empty string key (not found)",
        dict(
            obj=DictView({}),
            key='',
            exp_result=False,
        ),
        None, None, True
    ),
    (
        "Empty dict, with non-empty key (not found)",
        dict(
            obj=DictView({}),
            key='Dog',
            exp_result=False,
        ),
        None, None, True
    ),

    # Non-empty DictView
    (
        "Non-empty dict, with non-existing None key",
        dict(
            obj=DictView(OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')])),
            key=None,
            exp_result=False,
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with empty string key (not found)",
        dict(
            obj=DictView(OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')])),
            key='',
            exp_result=False,
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with non-empty key (not found)",
        dict(
            obj=DictView(OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')])),
            key='invalid',
            exp_result=False,
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with existing key",
        dict(
            obj=DictView(OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')])),
            key='Dog',
            exp_result=True,
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_CONTAINS)
@simplified_test_function
def test_DictView_contains(testcase, obj, key, exp_result):
    """
    Test function for DictView.__contains__() / key in ncd
    """

    # The code to be tested
    act_result = key in obj

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_result == exp_result, \
        "Unexpected result at key {k!r}".format(k=key)


TESTCASES_DICTVIEW_HAS_KEY = [

    # Testcases for DictView.has_key()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj: DictView object to be used for the test.
    #   * key: Key to be used for the test.
    #   * exp_result: Expected result (bool).
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty DictView
    (
        "Empty dict, with None key",
        dict(
            obj=DictView({}),
            key=None,
            exp_result=False,
        ),
        AttributeError if not DICT_SUPPORTS_HAS_KEY else None, None, True
    ),
    (
        "Empty dict, with integer key (no lower / success)",
        dict(
            obj=DictView({}),
            key=1234,
            exp_result=False,
        ),
        AttributeError if not DICT_SUPPORTS_HAS_KEY else None, None, True
    ),
    (
        "Empty dict, with empty string key (not found)",
        dict(
            obj=DictView({}),
            key='',
            exp_result=False,
        ),
        AttributeError if not DICT_SUPPORTS_HAS_KEY else None, None, True
    ),
    (
        "Empty dict, with non-empty key (not found)",
        dict(
            obj=DictView({}),
            key='Dog',
            exp_result=False,
        ),
        AttributeError if not DICT_SUPPORTS_HAS_KEY else None, None, True
    ),

    # Non-empty DictView
    (
        "Non-empty dict, with non-existing None key",
        dict(
            obj=DictView(OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')])),
            key=None,
            exp_result=False,
        ),
        AttributeError if not DICT_SUPPORTS_HAS_KEY else None, None, True
    ),
    (
        "Non-empty dict, with empty string key (not found)",
        dict(
            obj=DictView(OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')])),
            key='',
            exp_result=False,
        ),
        AttributeError if not DICT_SUPPORTS_HAS_KEY else None, None, True
    ),
    (
        "Non-empty dict, with non-empty key (not found)",
        dict(
            obj=DictView(OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')])),
            key='invalid',
            exp_result=False,
        ),
        AttributeError if not DICT_SUPPORTS_HAS_KEY else None, None, True
    ),
    (
        "Non-empty dict, with existing key",
        dict(
            obj=DictView(OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')])),
            key='Dog',
            exp_result=True,
        ),
        AttributeError if not DICT_SUPPORTS_HAS_KEY else None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_HAS_KEY)
@simplified_test_function
def test_DictView_has_key(testcase, obj, key, exp_result):
    """
    Test function for DictView.has_key()
    """

    # The code to be tested
    act_result = obj.has_key(key)  # noqa: W601

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_result == exp_result, \
        "Unexpected result at key {k!r}".format(k=key)


TESTCASES_DICTVIEW_REVERSED = [

    # Testcases for DictView.__reversed__() / reversed(ncd)

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj: DictView object to be used for the test.
    #   * exp_keys: Expected result as a list of keys, or None.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty dict",
        dict(
            obj=DictView({}),
            exp_keys=[],
        ),
        None if DICT_SUPPORTS_REVERSED else TypeError, None, DICT_IS_ORDERED
    ),
    (
        "Dict with one item",
        dict(
            obj=DictView({'Dog': 'Cat'}),
            exp_keys=['Dog'],
        ),
        None if DICT_SUPPORTS_REVERSED else TypeError, None, DICT_IS_ORDERED
    ),
    (
        "Dict with two items",
        dict(
            obj=DictView({'Dog': 'Cat', 'Budgie': 'Fish'}),
            exp_keys=['Budgie', 'Dog'],
        ),
        None if DICT_SUPPORTS_REVERSED else TypeError, None, DICT_IS_ORDERED
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_REVERSED)
@simplified_test_function
def test_DictView_reversed(testcase, obj, exp_keys):
    """
    Test function for DictView.__reversed__() / reversed(ncd)
    """

    # The code to be tested
    act_keys_iter = reversed(obj)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    # The reason we verify that an iterator is returned is that
    # DictView.__reversed__() delegates to keys() which returns a list in
    # Python 2, so this verifies that reversed() still turns this into an
    # iterator.
    assert isinstance(act_keys_iter, Iterator)

    act_keys = list(act_keys_iter)
    assert act_keys == exp_keys


TESTCASES_DICTVIEW_GET = [

    # Testcases for DictView.get()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj: DictView object to be used for the test.
    #   * key: Key to be used for the test.
    #   * default: Default value to be used for the test, or _OMIT_ARG.
    #   * exp_value: Expected value at the key.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty DictView
    (
        "Empty dict, with None key",
        dict(
            obj=DictView({}),
            key=None,
            default=_OMIT_ARG,
            exp_value=None,
        ),
        None, None, True
    ),
    (
        "Empty dict, with integer key",
        dict(
            obj=DictView({}),
            key=1234,
            default=_OMIT_ARG,
            exp_value=None,
        ),
        None, None, True
    ),
    (
        "Empty dict, with empty string key (defaulted without default)",
        dict(
            obj=DictView({}),
            key='',
            default=_OMIT_ARG,
            exp_value=None,
        ),
        None, None, True
    ),
    (
        "Empty dict, with empty string key (defaulted to a value)",
        dict(
            obj=DictView({}),
            key='',
            default='Newbie',
            exp_value='Newbie',
        ),
        None, None, True
    ),
    (
        "Empty dict, with non-empty key (defaulted without default)",
        dict(
            obj=DictView({}),
            key='Dog',
            default=_OMIT_ARG,
            exp_value=None,
        ),
        None, None, True
    ),
    (
        "Empty dict, with non-empty key (defaulted to a value)",
        dict(
            obj=DictView({}),
            key='Dog',
            default='Kitten',
            exp_value='Kitten',
        ),
        None, None, True
    ),

    # Non-empty DictView
    (
        "Non-empty dict, with None key",
        dict(
            obj=DictView(OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')])),
            key=None,
            default=_OMIT_ARG,
            exp_value=None,
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with empty string key (defaulted without default)",
        dict(
            obj=DictView(OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')])),
            key='',
            default=_OMIT_ARG,
            exp_value=None,
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with empty string key (defaulted to a value)",
        dict(
            obj=DictView(OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')])),
            key='',
            default='Newbie',
            exp_value='Newbie',
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with non-empty non-existing key (defaulted without "
        "default)",
        dict(
            obj=DictView(OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')])),
            key='invalid',
            default=_OMIT_ARG,
            exp_value=None,
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with non-empty non-existing key (defaulted to a "
        "value)",
        dict(
            obj=DictView(OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')])),
            key='invalid',
            default='Newbie',
            exp_value='Newbie',
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with existing key (no default)",
        dict(
            obj=DictView(OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')])),
            key='Dog',
            default=_OMIT_ARG,
            exp_value='Cat',
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_GET)
@simplified_test_function
def test_DictView_get(testcase, obj, key, default, exp_value):
    """
    Test function for DictView.get()
    """

    # The code to be tested
    if default is _OMIT_ARG:
        act_value = obj.get(key)
    else:
        act_value = obj.get(key, default)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_value == exp_value, \
        "Unexpected value at key {k!r} with default {d}". \
        format(k=key, d="omitted" if default is _OMIT_ARG else repr(default))


TESTCASES_DICTVIEW_ITEMS = [

    # Testcases for DictView.keys(), values(), items()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj: DictView object to be used for the test.
    #   * exp_items: List with expected items (key,value) in expected order.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty dict",
        dict(
            obj=DictView({}),
            exp_items=[],
        ),
        None, None, True
    ),
    (
        "Dict with one item",
        dict(
            obj=DictView(OrderedDict([('Dog', 'Cat')])),
            exp_items=[('Dog', 'Cat')],
        ),
        None, None, True
    ),
    (
        "Dict with two items",
        dict(
            obj=DictView(OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')])),
            exp_items=[('Dog', 'Cat'), ('Budgie', 'Fish')],
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_ITEMS)
@simplified_test_function
def test_DictView_keys(testcase, obj, exp_items):
    """
    Test function for DictView.keys()
    """

    # The code to be tested
    act_keys = obj.keys()

    # Also test iterating through the result
    act_keys_list = list(act_keys)

    # Test that second iteration is possible
    act_keys_list2 = list(act_keys)

    if not PY2:

        # Test __contained__() of the returned view
        for key in act_keys_list:
            assert key in act_keys

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    if PY2:
        assert isinstance(act_keys, list)
    else:
        assert isinstance(act_keys, KeysView)

    exp_keys = [item[0] for item in exp_items]
    assert act_keys_list == exp_keys
    assert act_keys_list2 == exp_keys


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_ITEMS)
@simplified_test_function
def test_DictView_values(testcase, obj, exp_items):
    """
    Test function for DictView.values()
    """

    # The code to be tested
    act_values = obj.values()

    # Also test iterating through the result
    act_values_list = list(act_values)

    # Test that second iteration is possible
    act_values_list2 = list(act_values)

    if not PY2:

        # Test __contained__() of the returned view
        for value in act_values_list:
            assert value in act_values

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    if PY2:
        assert isinstance(act_values, list)
    else:
        assert isinstance(act_values, ValuesView)

    exp_values = [item[1] for item in exp_items]
    assert act_values_list == exp_values
    assert act_values_list2 == exp_values


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_ITEMS)
@simplified_test_function
def test_DictView_items(testcase, obj, exp_items):
    """
    Test function for DictView.items()
    """

    # The code to be tested
    act_items = obj.items()

    # Also test iterating through the result
    act_items_list = list(act_items)

    # Test that second iteration is possible
    act_items_list2 = list(act_items)

    if not PY2:

        # Test __contained__() of the returned view
        for item in act_items_list:
            assert item in act_items

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    if PY2:
        assert isinstance(act_items, list)
    else:
        assert isinstance(act_items, ItemsView)

    assert act_items_list == exp_items
    assert act_items_list2 == exp_items


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_ITEMS)
@simplified_test_function
def test_DictView_iterkeys(testcase, obj, exp_items):
    """
    Test function for DictView.iterkeys()
    """

    if not DICT_SUPPORTS_ITER_VIEW:
        pytest.skip("Test dictionary does not support iterkeys() method")

    assert PY2

    # The code to be tested
    act_keys = []
    for key in obj.iterkeys():
        act_keys.append(key)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    exp_keys = [item[0] for item in exp_items]
    assert act_keys == exp_keys


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_ITEMS)
@simplified_test_function
def test_DictView_itervalues(testcase, obj, exp_items):
    """
    Test function for DictView.itervalues()
    """

    if not DICT_SUPPORTS_ITER_VIEW:
        pytest.skip("Test dictionary does not support itervalues() method")

    assert PY2

    # The code to be tested
    act_values = []
    for value in obj.itervalues():
        act_values.append(value)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    exp_values = [item[1] for item in exp_items]
    assert act_values == exp_values


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_ITEMS)
@simplified_test_function
def test_DictView_iteritems(testcase, obj, exp_items):
    """
    Test function for DictView.iteritemss()
    """

    if not DICT_SUPPORTS_ITER_VIEW:
        pytest.skip("Test dictionary does not support iteritems() method")

    assert PY2

    # The code to be tested
    act_items = []
    for item in obj.iteritems():
        act_items.append(item)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_items == exp_items


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_ITEMS)
@simplified_test_function
def test_DictView_viewkeys(testcase, obj, exp_items):
    """
    Test function for DictView.viewkeys() (PY2 only)
    """

    if not DICT_SUPPORTS_ITER_VIEW:
        pytest.skip("Test dictionary does support viewkeys() method")

    assert PY2

    # The code to be tested
    act_keys = obj.viewkeys()

    # Also test iterating through the result
    act_keys_list = list(act_keys)

    # Test that second iteration is possible
    act_keys_list2 = list(act_keys)

    # Test __contained__() of the returned view
    for key in act_keys_list:
        assert key in act_keys

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert isinstance(act_keys, KeysView)

    exp_keys = [item[0] for item in exp_items]
    assert act_keys_list == exp_keys
    assert act_keys_list2 == exp_keys


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_ITEMS)
@simplified_test_function
def test_DictView_viewvalues(testcase, obj, exp_items):
    """
    Test function for DictView.viewvalues()
    """

    if not DICT_SUPPORTS_ITER_VIEW:
        pytest.skip("Test dictionary does not support viewvalues() method")

    assert PY2

    # The code to be tested
    act_values = obj.viewvalues()

    # Also test iterating through the result
    act_values_list = list(act_values)

    # Test that second iteration is possible
    act_values_list2 = list(act_values)

    # Test __contained__() of the returned view
    for value in act_values_list:
        assert value in act_values

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert isinstance(act_values, ValuesView)

    exp_values = [item[1] for item in exp_items]
    assert act_values_list == exp_values
    assert act_values_list2 == exp_values


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_ITEMS)
@simplified_test_function
def test_DictView_viewitems(testcase, obj, exp_items):
    """
    Test function for DictView.viewitems()
    """

    if not DICT_SUPPORTS_ITER_VIEW:
        pytest.skip("Test dictionary does not support viewitems() method")

    assert PY2

    # The code to be tested
    act_items = obj.viewitems()

    # Also test iterating through the result
    act_items_list = list(act_items)

    # Test that second iteration is possible
    act_items_list2 = list(act_items)

    # Test __contained__() of the returned view
    for item in act_items_list:
        assert item in act_items

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert isinstance(act_items, ItemsView)

    assert act_items_list == exp_items
    assert act_items_list2 == exp_items


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_ITEMS)
@simplified_test_function
def test_DictView_iter(testcase, obj, exp_items):
    """
    Test function for DictView.__iter__() / for key in ncd
    """

    # The code to be tested
    act_keys = []
    for key in obj:
        act_keys.append(key)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    exp_keys = [item[0] for item in exp_items]
    assert act_keys == exp_keys


TESTCASES_DICTVIEW_REPR = [

    # Testcases for DictView.__repr__() / repr(ncd)

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj: DictView object to be used for the test.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty dict",
        dict(
            obj=DictView({}),
        ),
        None, None, True
    ),
    (
        "Dict with two items",
        dict(
            obj=DictView(OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')])),
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_REPR)
@simplified_test_function
def test_DictView_repr(testcase, obj):
    """
    Test function for DictView.__repr__() / repr(ncd)
    """

    # The code to be tested
    result = repr(obj)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert re.match(r'^DictView\(.*\)$', result)

    # Note: This only tests for existence of each item, not for excess items
    # or representing the correct order.
    for item in obj.items():
        exp_item_result1 = "{0!r}: {1!r}".format(*item)
        exp_item_result2 = "({0!r}, {1!r})".format(*item)
        assert exp_item_result1 in result or exp_item_result2 in result


TESTCASES_DICTVIEW_COPY = [

    # Testcases for DictView.copy()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * dictview: DictView object to be used for the test.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "DictView with empty dict",
        dict(
            dictview=DictView({}),
        ),
        None, None, True
    ),
    (
        "DictView with dict with two items",
        dict(
            dictview=DictView(
                dict([('Dog', 'Cat'), ('Budgie', 'Fish')])),
        ),
        None, None, True
    ),
    (
        "DictView with OrderedDict with two items",
        dict(
            dictview=DictView(
                OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')])),
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_COPY)
@simplified_test_function
def test_DictView_copy(testcase, dictview):
    """
    Test function for DictView.copy()
    """

    dictview_dict = dictview.dict

    # The code to be tested
    dictview_copy = dictview.copy()

    dictview_copy_dict = dictview_copy.dict

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    # Verify the result type
    # pylint: disable=unidiomatic-typecheck
    assert type(dictview_copy) is DictView

    # Verify the result is a different object than the DictView
    assert id(dictview_copy) != id(dictview)

    # Verify the new dictionary is a different object than the underlying
    # dictionary, if mutable
    if isinstance(dictview_dict, MutableMapping):
        assert id(dictview_copy_dict) != id(dictview_dict)

    # Verify the new dictionary has the same type as the underlying dictionary
    # pylint: disable=unidiomatic-typecheck
    assert type(dictview_copy_dict) == type(dictview_dict)

    # Verify the new dictionary is equal to the underlying dictionary
    assert dictview_copy_dict == dictview_dict


TESTCASES_DICTVIEW_EQUAL = [

    # Testcases for DictView.__eq__(), __ne__()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj1: DictView object #1 to use.
    #   * obj2: DictView object #2 to use.
    #   * exp_obj_equal: Expected equality of the objects.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty dictionary",
        dict(
            obj1=DictView({}),
            obj2=DictView({}),
            exp_obj_equal=True,
        ),
        None, None, True
    ),
    (
        "One item, keys and values equal",
        dict(
            obj1=DictView(OrderedDict([('k1', 'v1')])),
            obj2=DictView(OrderedDict([('k1', 'v1')])),
            exp_obj_equal=True,
        ),
        None, None, True
    ),
    (
        "One item, keys equal, values different",
        dict(
            obj1=DictView(OrderedDict([('k1', 'v1')])),
            obj2=DictView(OrderedDict([('k1', 'v1_x')])),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "One item, keys different, values equal",
        dict(
            obj1=DictView(OrderedDict([('k1', 'v1')])),
            obj2=DictView(OrderedDict([('k2', 'v1')])),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "One item, keys equal, values both None",
        dict(
            obj1=DictView(OrderedDict([('k1', None)])),
            obj2=DictView(OrderedDict([('k1', None)])),
            exp_obj_equal=True,
        ),
        None, None, True
    ),
    (
        "Two equal items, in same order",
        dict(
            obj1=DictView(OrderedDict([('k1', 'v1'), ('k2', 'v2')])),
            obj2=DictView(OrderedDict([('k1', 'v1'), ('k2', 'v2')])),
            exp_obj_equal=True,
        ),
        None, None, True
    ),
    (
        "Two equal items in ordered dict, in different order",
        dict(
            obj1=DictView(OrderedDict([('k1', 'v1'), ('k2', 'v2')])),
            obj2=DictView(OrderedDict([('k2', 'v2'), ('k1', 'v1')])),
            exp_obj_equal=False,
        ),
        None, None, False
    ),
    (
        "Two equal items in standard dict, in different order",
        dict(
            obj1=DictView(dict([('k1', 'v1'), ('k2', 'v2')])),
            obj2=DictView(dict([('k2', 'v2'), ('k1', 'v1')])),
            exp_obj_equal=not DICT_IS_ORDERED,
        ),
        None, None, False
    ),
    (
        "Comparing unicode value with bytes value",
        dict(
            obj1=DictView(OrderedDict([('k1', b'v1')])),
            obj2=DictView(OrderedDict([('k2', u'v2')])),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "Matching unicode key with string key",
        dict(
            obj1=DictView(OrderedDict([('k1', 'v1')])),
            obj2=DictView(OrderedDict([(u'k2', 'v2')])),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "Higher key missing",
        dict(
            obj1=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            obj2=DictView(OrderedDict([('Budgie', 'Fish')])),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "Lower key missing",
        dict(
            obj1=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            obj2=DictView(OrderedDict([('Dog', 'Cat')])),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "First non-matching key is less. But longer size!",
        dict(
            obj1=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            obj2=DictView(OrderedDict([
                ('Budgie', 'Fish'), ('Curly', 'Snake'), ('Cozy', 'Dog'),
            ])),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "Only non-matching keys that are less. But longer size!",
        dict(
            obj1=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            obj2=DictView(OrderedDict(
                [('Alf', 'F'), ('Anton', 'S'), ('Aussie', 'D')])),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "First non-matching key is greater. But shorter size!",
        dict(
            obj1=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            obj2=DictView(OrderedDict([('Budgio', 'Fish')])),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "Only non-matching keys that are greater. But shorter size!",
        dict(
            obj1=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            obj2=DictView(OrderedDict([('Zoe', 'F')])),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "Same size. First non-matching key is less",
        dict(
            obj1=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            obj2=DictView(OrderedDict(
                [('Budgie', 'Fish'), ('Curly', 'Snake')])),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "Same size. Only non-matching keys that are less",
        dict(
            obj1=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            obj2=DictView(OrderedDict([('Alf', 'F'), ('Anton', 'S')])),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "Same size. Only non-matching keys that are greater",
        dict(
            obj1=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            obj2=DictView(OrderedDict([('Zoe', 'F'), ('Zulu', 'S')])),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "Same size, only matching keys. First non-matching value is less",
        dict(
            obj1=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            obj2=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Car')])),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "Same size, only matching keys. First non-matching value is greater",
        dict(
            obj1=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            obj2=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Caz')])),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "A value raises TypeError when compared (and equal still succeeds)",
        dict(
            obj1=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            obj2=DictView(OrderedDict(
                [('Budgie', NonEquatable()), ('Dog', 'Cat')])),
            exp_obj_equal=False,
        ),
        TypeError, None, True
    ),

    (
        "Mixing types: DictView of dict and dict",
        dict(
            obj1=DictView(dict({'Dog': 'Cat'})),
            obj2=dict({'Dog': 'Cat'}),
            exp_obj_equal=True,
        ),
        None, None, True
    ),
    (
        "Mixing types: dict and DictView of dict",
        dict(
            obj1=dict({'Dog': 'Cat'}),
            obj2=DictView(dict({'Dog': 'Cat'})),
            exp_obj_equal=True,
        ),
        None, None, True
    ),
    (
        "Mixing types: DictView of dict and OrDict (not a dict)",
        dict(
            obj1=DictView(dict({'Dog': 'Cat'})),
            obj2=OrDict({'Dog': 'Cat'}),
            exp_obj_equal=True,
        ),
        None, None, True
    ),
    (
        "Mixing types: dict and DictView of OrDict (not a dict)",
        dict(
            obj1=dict({'Dog': 'Cat'}),
            obj2=DictView(OrDict({'Dog': 'Cat'})),
            exp_obj_equal=True,
        ),
        None, None, True
    ),
    (
        "Mixing types: DictView of OrDict (not a dict) and DictView of dict",
        dict(
            obj1=DictView(OrDict({'Dog': 'Cat'})),
            obj2=DictView(dict({'Dog': 'Cat'})),
            exp_obj_equal=True,
        ),
        None, None, True
    ),
    (
        "Mixing types: DictView of dict and DictView of OrDict (not a dict)",
        dict(
            obj1=DictView(dict({'Dog': 'Cat'})),
            obj2=DictView(OrDict({'Dog': 'Cat'})),
            exp_obj_equal=True,
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_EQUAL)
@simplified_test_function
def test_DictView_eq(testcase, obj1, obj2, exp_obj_equal):
    """
    Test function for DictView.__eq__() / ncd1==ncd2
    """

    # Double check they are different objects
    assert id(obj1) != id(obj2)

    # The code to be tested
    eq1 = (obj1 == obj2)
    eq2 = (obj2 == obj1)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert eq1 == exp_obj_equal
    assert eq2 == exp_obj_equal


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_EQUAL)
@simplified_test_function
def test_DictView_ne(testcase, obj1, obj2, exp_obj_equal):
    """
    Test function for DictView.__ne__() / ncd1!=ncd2
    """

    # Double check they are different objects
    assert id(obj1) != id(obj2)

    # The code to be tested
    ne1 = (obj1 != obj2)
    ne2 = (obj2 != obj1)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert ne1 != exp_obj_equal
    assert ne2 != exp_obj_equal


TESTCASES_DICTVIEW_ORDERING = [

    # Testcases for DictView.__le__(), __lt__(), __ge__(), __gt__() / ord.ops

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj1: DictView object #1 to be used.
    #   * obj2: DictView object #2 to be used.
    #   * op: Order comparison operator to be used, as a string (e.g. '>')
    #   * exp_result: Expected result of the comparison, or None.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty dicts
    (
        "Empty dicts with >",
        dict(
            obj1=DictView({}),
            obj2=DictView({}),
            op='>',
            exp_result=False,
        ),
        None if DICT_SUPPORTS_COMPARISON else TypeError, None, True
    ),
    (
        "Empty dicts with >=",
        dict(
            obj1=DictView({}),
            obj2=DictView({}),
            op='>=',
            exp_result=True,
        ),
        None if DICT_SUPPORTS_COMPARISON else TypeError, None, True
    ),
    (
        "Empty dicts with <",
        dict(
            obj1=DictView({}),
            obj2=DictView({}),
            op='<',
            exp_result=False,
        ),
        None if DICT_SUPPORTS_COMPARISON else TypeError, None, True
    ),
    (
        "Empty dicts with <=",
        dict(
            obj1=DictView({}),
            obj2=DictView({}),
            op='<=',
            exp_result=True,
        ),
        None if DICT_SUPPORTS_COMPARISON else TypeError, None, True
    ),

    # Equal dicts
    (
        "Equal dicts with >",
        dict(
            obj1=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            obj2=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            op='>',
            exp_result=False,
        ),
        None if DICT_SUPPORTS_COMPARISON else TypeError, None, True
    ),
    (
        "Equal dicts with >=",
        dict(
            obj1=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            obj2=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            op='>=',
            exp_result=True,
        ),
        None if DICT_SUPPORTS_COMPARISON else TypeError, None, True
    ),
    (
        "Equal dicts with <",
        dict(
            obj1=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            obj2=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            op='<',
            exp_result=False,
        ),
        None if DICT_SUPPORTS_COMPARISON else TypeError, None, True
    ),
    (
        "Equal dicts with <=",
        dict(
            obj1=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            obj2=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            op='<=',
            exp_result=True,
        ),
        None if DICT_SUPPORTS_COMPARISON else TypeError, None, True
    ),

    # Dicts that compare less (obj1 < obj2)
    (
        "Less-comparing dicts with >",
        dict(
            obj1=DictView(OrderedDict([('Budgie', 'Fish')])),
            obj2=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            op='>',
            exp_result=False,
        ),
        None if DICT_SUPPORTS_COMPARISON else TypeError, None, True
    ),
    (
        "Less-comparing dicts with >=",
        dict(
            obj1=DictView(OrderedDict([('Budgie', 'Fish')])),
            obj2=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            op='>=',
            exp_result=False,
        ),
        None if DICT_SUPPORTS_COMPARISON else TypeError, None, True
    ),
    (
        "Less-comparing dicts with <",
        dict(
            obj1=DictView(OrderedDict([('Budgie', 'Fish')])),
            obj2=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            op='<',
            exp_result=True,
        ),
        None if DICT_SUPPORTS_COMPARISON else TypeError, None, True
    ),
    (
        "Less-comparing dicts with <=",
        dict(
            obj1=DictView(OrderedDict([('Budgie', 'Fish')])),
            obj2=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            op='<=',
            exp_result=True,
        ),
        None if DICT_SUPPORTS_COMPARISON else TypeError, None, True
    ),

    # Dicts that compare greater (obj1 > obj2)
    (
        "Greater-comparing dicts with >",
        dict(
            obj1=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            obj2=DictView(OrderedDict([('Budgie', 'Fish')])),
            op='>',
            exp_result=True,
        ),
        None if DICT_SUPPORTS_COMPARISON else TypeError, None, True
    ),
    (
        "Greater-comparing dicts with >=",
        dict(
            obj1=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            obj2=DictView(OrderedDict([('Budgie', 'Fish')])),
            op='>=',
            exp_result=True,
        ),
        None if DICT_SUPPORTS_COMPARISON else TypeError, None, True
    ),
    (
        "Greater-comparing dicts with <",
        dict(
            obj1=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            obj2=DictView(OrderedDict([('Budgie', 'Fish')])),
            op='<',
            exp_result=False,
        ),
        None if DICT_SUPPORTS_COMPARISON else TypeError, None, True
    ),
    (
        "Greater-comparing dicts with <=",
        dict(
            obj1=DictView(OrderedDict([('Budgie', 'Fish'), ('Dog', 'Cat')])),
            obj2=DictView(OrderedDict([('Budgie', 'Fish')])),
            op='<=',
            exp_result=False,
        ),
        None if DICT_SUPPORTS_COMPARISON else TypeError, None, True
    ),

    # Note: More subtle cases of less- or greater-comparing dicts are not
    # tested because the ordering comparison for DictView is deprecated.
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_ORDERING)
@simplified_test_function
def test_DictView_ordering(testcase, obj1, obj2, op, exp_result):
    """
    Test function for DictView.__le__(), __lt__(), __ge__(), __gt__() / ord.
    """

    comp_str = 'obj1 %s obj2' % op

    # Double check they are different objects
    assert id(obj1) != id(obj2)

    # The code to be tested
    result = eval(comp_str)  # pylint: disable=eval-used

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert result == exp_result


TESTCASES_DICTVIEW_PICKLE = [

    # Testcases for pickling and unpickling DictView objects

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * dictview: DictView object to be used for the test.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "DictView of empty dict",
        dict(
            dictview=DictView({}),
        ),
        None, None, True
    ),
    (
        "DictView of dict with one item",
        dict(
            dictview=DictView({'Dog': 'Cat'}),
        ),
        None, None, True
    ),
    (
        "DictView of DerivedDict with one item",
        dict(
            dictview=DictView(DerivedDict({'Dog': 'Cat'})),
        ),
        None, None, True
    ),
    (
        "DerivedDictView of dict with one item",
        dict(
            dictview=DerivedDictView(dict({'Dog': 'Cat'})),
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_PICKLE)
@simplified_test_function
def test_DictView_pickle(testcase, dictview):
    """
    Test function for pickling and unpickling DictView objects
    """

    # Pickle the object
    pkl = pickle.dumps(dictview)

    # Unpickle the object
    dictview2 = pickle.loads(pkl)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert dictview2 == dictview


TESTCASES_DICTVIEW_HASH = [

    # Testcases for DictView.__hash__()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * dict_obj: Underlying dictionary object.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty dict",
        dict(
            dict_obj={},
        ),
        TypeError, None, True
    ),
    (
        "Empty OrderedDict",
        dict(
            dict_obj=OrderedDict(),
        ),
        TypeError, None, True
    ),
    (
        "Empty HashableDict",
        dict(
            dict_obj=HashableDict(),
        ),
        None, None, True
    ),

    (
        "Dict with one item",
        dict(
            dict_obj=dict([('a', 1)]),
        ),
        TypeError, None, True
    ),
    (
        "OrderedDict with one item",
        dict(
            dict_obj=OrderedDict([('a', 1)]),
        ),
        TypeError, None, True
    ),
    (
        "HashableDict with one item",
        dict(
            dict_obj=HashableDict([('a', 1)]),
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_HASH)
@simplified_test_function
def test_DictView_hash(testcase, dict_obj):
    """
    Test function for DictView.__hash__() / hash()
    """

    view = DictView(dict_obj)

    # The code to be tested
    view_hash = hash(view)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    # If it worked for the view, it must work for the underlying collection
    dict_hash = hash(dict_obj)

    # Verify the hash value of the underlying collection is used for the view
    assert view_hash == dict_hash


TESTCASES_DICTVIEW_OR = [

    # Testcases for DictView.__or__/__ror__() / self | or

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj1: DictView object #1 to be used.
    #   * obj2: DictView object #2 to be used.
    #   * exp_result: Expected result of the operation, or None.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "DictView of empty NorDict and DictView of empty NorDict",
        dict(
            obj1=DictView(NorDict()),
            obj2=DictView(NorDict()),
            exp_result=None,
        ),
        TypeError, None, True
    ),
    (
        "DictView of empty OrDict and DictView of empty NorDict",
        dict(
            obj1=DictView(OrDict()),
            obj2=DictView(NorDict()),
            exp_result=DictView(OrDict()),
        ),
        None if DICT_SUPPORTS_OR else TypeError, None, True
    ),
    (
        "DictView of empty NorDict and DictView of empty OrDict",
        dict(
            obj1=DictView(NorDict()),
            obj2=DictView(OrDict()),
            exp_result=DictView(NorDict()),
        ),
        None if DICT_SUPPORTS_OR else TypeError, None, True
    ),
    (
        "Empty NorDict and DictView of empty OrDict",
        dict(
            obj1=NorDict(),
            obj2=DictView(OrDict()),
            exp_result=NorDict(),
        ),
        None if DICT_SUPPORTS_OR else TypeError, None, True
    ),

    (
        "OrDict with two items and NorDict with equal two items",
        dict(
            obj1=DictView(OrDict({'Budgie': 'Fish', 'Dog': 'Cat'})),
            obj2=DictView(NorDict({'Budgie': 'Fish', 'Dog': 'Cat'})),
            exp_result=DictView(OrDict({'Budgie': 'Fish', 'Dog': 'Cat'})),
        ),
        None if DICT_SUPPORTS_OR else TypeError, None, True
    ),
    (
        "NorDict with two items and OrDict with equal two items",
        dict(
            obj1=DictView(NorDict({'Budgie': 'Fish', 'Dog': 'Cat'})),
            obj2=DictView(OrDict({'Budgie': 'Fish', 'Dog': 'Cat'})),
            exp_result=DictView(NorDict({'Budgie': 'Fish', 'Dog': 'Cat'})),
        ),
        None if DICT_SUPPORTS_OR else TypeError, None, True
    ),

    (
        "Empty OrDict and NorDict with two items",
        dict(
            obj1=DictView(OrDict()),
            obj2=DictView(NorDict({'Budgie': 'Fish', 'Dog': 'Cat'})),
            exp_result=DictView(OrDict({'Budgie': 'Fish', 'Dog': 'Cat'})),
        ),
        None if DICT_SUPPORTS_OR else TypeError, None, True
    ),
    (
        "Empty NorDict and OrDict with two items",
        dict(
            obj1=DictView(NorDict()),
            obj2=DictView(OrDict({'Budgie': 'Fish', 'Dog': 'Cat'})),
            exp_result=DictView(NorDict({'Budgie': 'Fish', 'Dog': 'Cat'})),
        ),
        None if DICT_SUPPORTS_OR else TypeError, None, True
    ),

    (
        "OrDict with two items and empty NorDict",
        dict(
            obj1=DictView(OrDict({'Budgie': 'Fish', 'Dog': 'Cat'})),
            obj2=DictView(NorDict()),
            exp_result=DictView(OrDict({'Budgie': 'Fish', 'Dog': 'Cat'})),
        ),
        None if DICT_SUPPORTS_OR else TypeError, None, True
    ),
    (
        "NorDict with two items and empty OrDict",
        dict(
            obj1=DictView(NorDict({'Budgie': 'Fish', 'Dog': 'Cat'})),
            obj2=DictView(OrDict()),
            exp_result=DictView(NorDict({'Budgie': 'Fish', 'Dog': 'Cat'})),
        ),
        None if DICT_SUPPORTS_OR else TypeError, None, True
    ),

    (
        "OrDict with one item and NorDict with one item with different key",
        dict(
            obj1=DictView(OrDict({'Budgie': 'Fish'})),
            obj2=DictView(NorDict({'Dog': 'Cat'})),
            exp_result=DictView(OrDict({'Budgie': 'Fish', 'Dog': 'Cat'})),
        ),
        None if DICT_SUPPORTS_OR else TypeError, None, True
    ),
    (
        "NorDict with one item and OrDict with one item with different key",
        dict(
            obj1=DictView(NorDict({'Budgie': 'Fish'})),
            obj2=DictView(OrDict({'Dog': 'Cat'})),
            exp_result=DictView(NorDict({'Budgie': 'Fish', 'Dog': 'Cat'})),
        ),
        None if DICT_SUPPORTS_OR else TypeError, None, True
    ),

    (
        "OrDict with one item and NorDict with one item with same key, "
        "different value",
        dict(
            obj1=DictView(OrDict({'Budgie': 'Fish'})),
            obj2=DictView(NorDict({'Budgie': 'Cat'})),
            exp_result=DictView(NorDict({'Budgie': 'Cat'})),
        ),
        None if DICT_SUPPORTS_OR else TypeError, None, True
    ),
    (
        "NorDict with one item and OrDict with one item with same key, "
        "different value",
        dict(
            obj1=DictView(NorDict({'Budgie': 'Fish'})),
            obj2=DictView(OrDict({'Budgie': 'Cat'})),
            exp_result=DictView(NorDict({'Budgie': 'Cat'})),
        ),
        None if DICT_SUPPORTS_OR else TypeError, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DICTVIEW_OR)
@simplified_test_function
def test_DictView_or(testcase, obj1, obj2, exp_result):
    """
    Test function for DictView.__or__/__ror__() / self | or
    """

    # The code to be tested
    result = obj1 | obj2

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert result == exp_result



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
immutable-views - Immutable views on other collection objects
"""

from __future__ import absolute_import

from ._dict_view import *  # noqa: F403,F401
from ._list_view import *  # noqa: F403,F401
from ._set_view import *  # noqa: F403,F401
from . import _version

#: The full version of this package including any development levels, as a
#: :term:`string`.
#:
#: Possible formats for this version string are:
#:
#: * "M.N.P.dev1": Development level 1 of a not yet released version M.N.P
#: * "M.N.P": A released version M.N.P
__version__ = _version.__version__
